# Copyright (C) 2020 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GPL v3 or later

APP = 'git-delete-merged-branches'
DESCRIPTION = 'Command-line tool to delete merged Git branches'
VERSION = '1.1.2'
